/*
@license
The following license applies to all parts of this software except as
documented below.

    Copyright (c) 2019, Twilio, inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in
         the documentation and/or other materials provided with the
         distribution.

      3. Neither the name of Twilio nor the names of its contributors may
         be used to endorse or promote products derived from this software
         without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software includes javascript-state-machine under the following license.

    Copyright (c) 2012, 2013, 2014, 2015, Jake Gordon and contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE

This software includes loglevel under the following license.

    Copyright (c) 2013 Tim Perry

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

This software includes q under the following license.

    Copyright 2009–2014 Kristopher Michael Kowal. All rights reserved.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

This software includes platform.js under the following license.

    Copyright 2014 Benjamin Tan <https://d10.github.io/>
    Copyright 2011-2015 John-David Dalton <http://allyoucanleet.com/>

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var logger = require('../logger.js');
var channel = require('../channel.js');
var channeldescriptor = require('../channeldescriptor.js');
var deferred = require('../util/deferred.js');
var index = require('../util/index.js');
var isEqual = require('lodash.isequal');
var replayEventEmitter = require('@twilio/replay-event-emitter');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var isEqual__default = /*#__PURE__*/_interopDefaultLegacy(isEqual);

const log = logger.Logger.scope('Channels');
/**
 * Represents channels collection
 * {@see Channel}
 */
class Channels extends replayEventEmitter.ReplayEventEmitter {
    constructor(configuration, services) {
        super();
        this.configuration = configuration;
        this.services = services;
        this.channels = new Map();
        this.tombstones = new Set();
        this.myChannelsFetched = false;
        this.myChannelsRead = new deferred.Deferred();
    }
    async getMap() {
        return await this.services.syncClient.map({
            id: this.configuration.myConversations,
            mode: 'open_existing'
        });
    }
    /**
     * Add channel to server
     * @private
     * @returns {Promise<Channel>} Channel
     */
    async addChannel(options) {
        let attributes;
        if (typeof options.attributes === 'undefined') {
            attributes = {};
        }
        else {
            attributes = options.attributes;
        }
        const response = await this.services.commandExecutor.mutateResource('post', this.configuration.links.conversations, {
            type: options.isPrivate ? 'private' : 'public',
            unique_name: options.uniqueName,
            friendly_name: options.friendlyName,
            attributes: attributes !== undefined ? JSON.stringify(attributes) : undefined,
        });
        const channelSid = response.sid || null;
        const channelDocument = response.sync_objects.conversation || null;
        const links = Object.assign({ self: response.url }, response.links);
        let existingChannel = this.channels.get(channelSid);
        if (existingChannel) {
            await existingChannel._subscribe();
            return existingChannel;
        }
        let channel$1 = new channel.Channel({
            channel: channelDocument,
            entityName: null,
            uniqueName: null,
            attributes: null,
            createdBy: null,
            friendlyName: null,
            lastConsumedMessageIndex: null,
            type: options.isPrivate ? 'private' : 'public',
            dateCreated: null,
            dateUpdated: null
        }, channelSid, links, this.configuration, this.services);
        this.channels.set(channel$1.sid, channel$1);
        this.registerForEvents(channel$1);
        await channel$1._subscribe();
        this.emit('channelAdded', channel$1);
        return channel$1;
    }
    /**
     * Fetch channels list and instantiate all necessary objects
     */
    async fetchChannels() {
        try {
            const map = await this.getMap();
            map.on('itemAdded', args => {
                log.debug(`itemAdded: ${args.item.key}`);
                this.upsertChannel('sync', args.item.key, args.item.data);
            });
            map.on('itemRemoved', args => {
                log.debug(`itemRemoved: ${args.key}`);
                const sid = args.key;
                if (!this.myChannelsFetched) {
                    this.tombstones.add(sid);
                }
                const channel = this.channels.get(sid);
                if (!channel) {
                    return;
                }
                if (channel.status === 'joined' || channel.status === 'invited') {
                    channel._setStatus('notParticipating', 'sync');
                    this.emit('channelLeft', channel);
                }
                if (channel.isPrivate) {
                    this.channels.delete(sid);
                    this.emit('channelRemoved', channel);
                    channel.emit('removed', channel);
                }
            });
            map.on('itemUpdated', args => {
                log.debug(`itemUpdated: ${args.item.key}`);
                this.upsertChannel('sync', args.item.key, args.item.data);
            });
            const myChannels = await this._fetchMyChannels();
            const upserts = [];
            for (const channel of myChannels) {
                upserts.push(this.upsertChannel('rest', channel.channel_sid, channel));
            }
            this.myChannelsRead.set(true);
            await Promise.all(upserts);
            this.myChannelsFetched = true;
            this.tombstones.clear();
            log.debug('The channels list has been successfully fetched');
            return this;
        }
        catch (error) {
            const errorMessage = 'Failed to fetch the channels list';
            if (this.services.syncClient.connectionState !== 'disconnected') {
                log.error(errorMessage, error);
            }
            log.debug(`ERROR: ${errorMessage}`, error);
            throw error;
        }
    }
    async _wrapPaginator(page, op) {
        return op(page.items)
            .then(items => ({
            items: items,
            hasNextPage: page.hasNextPage,
            hasPrevPage: page.hasPrevPage,
            nextPage: () => page.nextPage().then(x => this._wrapPaginator(x, op)),
            prevPage: () => page.prevPage().then(x => this._wrapPaginator(x, op))
        }));
    }
    async getChannels(args) {
        const channelsMap = await this.getMap();
        const page = await channelsMap.getItems(args);
        return this._wrapPaginator(page, items => Promise.all(items.map((item) => this.upsertChannel('sync', item.key, item.data))));
    }
    async getChannel(sid) {
        const channelsMap = await this.getMap();
        const page = await channelsMap.getItems({ key: sid });
        const items = page.items.map((item) => this.upsertChannel('sync', item.key, item.data));
        return items.length > 0 ? items[0] : null;
    }
    async pushChannel(descriptor) {
        const sid = descriptor.sid;
        const data = {
            entityName: null,
            lastConsumedMessageIndex: descriptor.lastConsumedMessageIndex,
            type: descriptor.type,
            status: descriptor.status,
            friendlyName: descriptor.friendlyName,
            dateUpdated: descriptor.dateUpdated,
            dateCreated: descriptor.dateCreated,
            uniqueName: descriptor.uniqueName,
            createdBy: descriptor.createdBy,
            attributes: descriptor.attributes,
            channel: descriptor instanceof channeldescriptor.ChannelDescriptor ? descriptor.channel : null,
            notificationLevel: descriptor.notificationLevel,
            sid: sid
        };
        return await this.upsertChannel('chat', sid, data);
    }
    _updateChannel(source, channel, data) {
        const areSourcesDifferent = channel._statusSource() !== undefined && source !== channel._statusSource();
        const isChannelSourceSync = source !== 'rest' || channel._statusSource() === 'sync';
        if (areSourcesDifferent && isChannelSourceSync && source !== 'sync') {
            log.trace('upsertChannel: the channel is known from sync and it came from chat, ignoring', {
                sid: channel.sid,
                data: data.status,
                channel: channel.status
            });
            return;
        }
        if (['joined', 'invited'].includes(data.status) && channel.status !== data.status) {
            channel._setStatus(data.status, source);
            let updateData = {};
            if (data.notificationLevel !== undefined) {
                updateData.notificationLevel = data.notificationLevel;
            }
            if (data.lastConsumedMessageIndex !== undefined) {
                updateData.lastConsumedMessageIndex = data.lastConsumedMessageIndex;
            }
            if (!isEqual__default['default'](updateData, {})) {
                channel._update(updateData);
            }
            channel._subscribe().then(() => {
                this.emit(data.status === 'joined' ? 'channelJoined' : 'channelInvited', channel);
            });
            return;
        }
        if (['joined', 'invited'].includes(channel.status) && data.status === 'notParticipating') {
            channel._setStatus('notParticipating', source);
            channel._update(data);
            channel._subscribe().then(() => {
                this.emit('channelLeft', channel);
            });
            return;
        }
        if (data.type === 'private' && data.status === 'notParticipating') {
            channel._subscribe();
            return;
        }
        channel._update(data);
    }
    async upsertChannel(source, sid, data) {
        log.trace(`upsertChannel called for ${sid}`, data);
        const channel$1 = this.channels.get(sid);
        // If the channel is known, update it
        if (channel$1) {
            log.trace(`upsertChannel: the channel ${channel$1.sid} is known;` +
                `its status is known from source ${channel$1._statusSource()} ` +
                `and the update came from source ${source}`, channel$1);
            this._updateChannel(source, channel$1, data);
            await channel$1._subscribe();
            return channel$1;
        }
        // If the channel is deleted, ignore it
        if (['chat', 'rest'].includes(source) && this.tombstones.has(sid)) {
            log.trace('upsertChannel: the channel is deleted but reappeared again from chat, ignoring', sid);
            return null;
        }
        // If the channel is unknown, fetch it
        log.trace(`upsertChannel: creating a local channel object with sid ${sid}`, data);
        const baseLink = `${this.configuration.links.conversations}/${sid}`;
        const links = {
            self: baseLink,
            messages: `${baseLink}/Messages`,
            participants: `${baseLink}/Participants`,
            invites: `${baseLink}/Invites`
        };
        const newChannel = new channel.Channel(data, sid, links, this.configuration, this.services);
        this.channels.set(sid, newChannel);
        await newChannel._subscribe();
        this.registerForEvents(newChannel);
        this.emit('channelAdded', newChannel);
        if (['joined', 'invited'].includes(data.status)) {
            newChannel._setStatus(data.status, source);
            this.emit(data.status === 'joined' ? 'channelJoined' : 'channelInvited', newChannel);
        }
        return newChannel;
    }
    onChannelRemoved(sid) {
        let channel = this.channels.get(sid);
        if (channel) {
            this.channels.delete(sid);
            this.emit('channelRemoved', channel);
        }
    }
    registerForEvents(channel) {
        channel.on('removed', () => this.onChannelRemoved(channel.sid));
        channel.on('updated', (args) => this.emit('channelUpdated', args));
        channel.on('memberJoined', this.emit.bind(this, 'memberJoined'));
        channel.on('memberLeft', this.emit.bind(this, 'memberLeft'));
        channel.on('memberUpdated', (args) => this.emit('memberUpdated', args));
        channel.on('messageAdded', this.emit.bind(this, 'messageAdded'));
        channel.on('messageUpdated', (args) => this.emit('messageUpdated', args));
        channel.on('messageRemoved', this.emit.bind(this, 'messageRemoved'));
        channel.on('typingStarted', this.emit.bind(this, 'typingStarted'));
        channel.on('typingEnded', this.emit.bind(this, 'typingEnded'));
    }
    async _fetchMyChannels() {
        let channels = [];
        let pageToken = null;
        do {
            const url = new index.UriBuilder(this.configuration.links.myConversations);
            if (pageToken) {
                url.arg('PageToken', pageToken);
            }
            const response = await this.services.network.get(url.build());
            const preProcessedChannels = response.body.conversations.map((channelDescriptor) => ({
                descriptor: channelDescriptor,
                channel_sid: channelDescriptor.conversation_sid,
                status: channelDescriptor.status,
                channel: channelDescriptor.sync_objects.conversation,
                messages: channelDescriptor.sync_objects.messages,
                roster: `${channelDescriptor.conversation_sid}.roster`,
                lastConsumedMessageIndex: channelDescriptor.last_consumed_message_index,
                notificationLevel: channelDescriptor.notification_level
            }));
            pageToken = response.body.meta.next_token;
            channels = [...channels, ...preProcessedChannels];
        } while (pageToken);
        return channels;
    }
}

exports.Channel = channel.Channel;
exports.Channels = Channels;
//# sourceMappingURL=channels.js.map
