/*
@license
The following license applies to all parts of this software except as
documented below.

    Copyright (c) 2019, Twilio, inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in
         the documentation and/or other materials provided with the
         distribution.

      3. Neither the name of Twilio nor the names of its contributors may
         be used to endorse or promote products derived from this software
         without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software includes javascript-state-machine under the following license.

    Copyright (c) 2012, 2013, 2014, 2015, Jake Gordon and contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE

This software includes loglevel under the following license.

    Copyright (c) 2013 Tim Perry

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

This software includes q under the following license.

    Copyright 2009–2014 Kristopher Michael Kowal. All rights reserved.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

This software includes platform.js under the following license.

    Copyright 2014 Benjamin Tan <https://d10.github.io/>
    Copyright 2011-2015 John-David Dalton <http://allyoucanleet.com/>

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var member = require('../member.js');
var logger = require('../logger.js');
var replayEventEmitter = require('@twilio/replay-event-emitter');

const log = logger.Logger.scope('Members');
/**
 * @classdesc Represents the collection of members for the channel
 * @fires Members#memberJoined
 * @fires Members#memberLeft
 * @fires Members#memberUpdated
 */
class Members extends replayEventEmitter.ReplayEventEmitter {
    constructor(channel, members, links, configuration, services) {
        super();
        this.channel = channel;
        this.members = members;
        this.links = links;
        this.configuration = configuration;
        this.services = services;
    }
    async unsubscribe() {
        if (this.rosterEntityPromise) {
            let entity = await this.rosterEntityPromise;
            entity.close();
            this.rosterEntityPromise = null;
        }
    }
    async subscribe(rosterObjectName) {
        return this.rosterEntityPromise = this.rosterEntityPromise
            || this.services.syncClient.map({ id: rosterObjectName, mode: 'open_existing' })
                .then(rosterMap => {
                rosterMap.on('itemAdded', args => {
                    log.debug(this.channel.sid + ' itemAdded: ' + args.item.key);
                    this.upsertMember(args.item.key, args.item.data)
                        .then(member => {
                        this.emit('memberJoined', member);
                    });
                });
                rosterMap.on('itemRemoved', args => {
                    log.debug(this.channel.sid + ' itemRemoved: ' + args.key);
                    let memberSid = args.key;
                    if (!this.members.has(memberSid)) {
                        return;
                    }
                    let leftMember = this.members.get(memberSid);
                    this.members.delete(memberSid);
                    this.emit('memberLeft', leftMember);
                });
                rosterMap.on('itemUpdated', args => {
                    log.debug(this.channel.sid + ' itemUpdated: ' + args.item.key);
                    this.upsertMember(args.item.key, args.item.data);
                });
                let membersPromises = [];
                let that = this;
                const rosterMapHandler = function (paginator) {
                    paginator.items.forEach(item => { membersPromises.push(that.upsertMember(item.key, item.data)); });
                    return paginator.hasNextPage ? paginator.nextPage().then(rosterMapHandler) : null;
                };
                return rosterMap
                    .getItems()
                    .then(rosterMapHandler)
                    .then(() => Promise.all(membersPromises))
                    .then(() => rosterMap);
            })
                .catch(err => {
                this.rosterEntityPromise = null;
                if (this.services.syncClient.connectionState != 'disconnected') {
                    log.error('Failed to get roster object for channel', this.channel.sid, err);
                }
                log.debug('ERROR: Failed to get roster object for channel', this.channel.sid, err);
                throw err;
            });
    }
    async upsertMember(memberSid, data) {
        let member$1 = this.members.get(memberSid);
        if (member$1) {
            return member$1._update(data);
        }
        const links = {
            self: `${this.links.participants}/${memberSid}`
        };
        member$1 = new member.Member(data, memberSid, this.channel, links, this.services);
        this.members.set(memberSid, member$1);
        member$1.on('updated', (args) => this.emit('memberUpdated', args));
        return member$1;
    }
    /**
     * @returns {Promise<Array<Member>>} returns list of members {@see Member}
     */
    async getMembers() {
        return this.rosterEntityPromise.then(() => {
            let members = [];
            this.members.forEach((member) => members.push(member));
            return members;
        });
    }
    /**
     * Get member by SID from channel
     * @returns {Promise<Member>}
     */
    async getMemberBySid(memberSid) {
        return this.rosterEntityPromise.then(() => {
            let member = this.members.get(memberSid);
            if (!member) {
                throw new Error('Member with SID ' + memberSid + ' was not found');
            }
            return member;
        });
    }
    /**
     * Get member by identity from channel
     * @returns {Promise<Member>}
     */
    async getMemberByIdentity(identity) {
        let foundMember = null;
        return this.rosterEntityPromise.then(() => {
            this.members.forEach((member) => {
                if (member.identity === identity) {
                    foundMember = member;
                }
            });
            if (!foundMember) {
                throw new Error('Member with identity ' + identity + ' was not found');
            }
            return foundMember;
        });
    }
    /**
     * Add user to the channel
     * @returns {Promise<any>}
     */
    async add(identity) {
        return await this.services.commandExecutor.mutateResource('post', this.links.participants, {
            identity
        });
    }
    /**
     * Invites user to the channel
     * User can choose either to join or not
     * @returns {Promise<any>}
     */
    async invite(identity) {
        return await this.services.commandExecutor.mutateResource('post', this.channel.links.invites, {
            identity
        });
    }
    /**
     * Remove member from channel
     * @returns {Promise<any>}
     */
    async remove(identity) {
        return await this.services.commandExecutor.mutateResource('delete', `${this.links.participants}/${identity}`);
    }
}
/**
 * Fired when member joined channel
 * @event Members#memberJoined
 * @type {Member}
 */
/**
 * Fired when member left channel
 * @event Members#memberLeft
 * @type {Member}
 */
/**
 * Fired when member updated
 * @event Members#memberUpdated
 * @type {Object}
 * @property {Member} member - Updated Member
 * @property {Member#UpdateReason[]} updateReasons - Array of Member's updated event reasons
 */

exports.Members = Members;
//# sourceMappingURL=members.js.map
