/*
@license
The following license applies to all parts of this software except as
documented below.

    Copyright (c) 2019, Twilio, inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in
         the documentation and/or other materials provided with the
         distribution.

      3. Neither the name of Twilio nor the names of its contributors may
         be used to endorse or promote products derived from this software
         without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software includes javascript-state-machine under the following license.

    Copyright (c) 2012, 2013, 2014, 2015, Jake Gordon and contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE

This software includes loglevel under the following license.

    Copyright (c) 2013 Tim Perry

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

This software includes q under the following license.

    Copyright 2009–2014 Kristopher Michael Kowal. All rights reserved.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

This software includes platform.js under the following license.

    Copyright 2014 Benjamin Tan <https://d10.github.io/>
    Copyright 2011-2015 John-David Dalton <http://allyoucanleet.com/>

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var logger = require('../logger.js');
var message = require('../message.js');
var replayEventEmitter = require('@twilio/replay-event-emitter');

const log = logger.Logger.scope('Messages');
/**
 * Represents the collection of messages in a channel
 */
class Messages extends replayEventEmitter.ReplayEventEmitter {
    constructor(channel, configuration, services) {
        super();
        this.channel = channel;
        this.configuration = configuration;
        this.services = services;
        this.messagesByIndex = new Map();
        this.messagesListPromise = null;
    }
    /**
     * Subscribe to the Messages Event Stream
     * @param {String} name - The name of Sync object for the Messages resource.
     * @returns {Promise}
     */
    async subscribe(name) {
        return this.messagesListPromise =
            this.messagesListPromise ||
                this.services.syncClient.list({ id: name, mode: 'open_existing' })
                    .then(list => {
                    list.on('itemAdded', args => {
                        log.debug(this.channel.sid + ' itemAdded: ' + args.item.index);
                        const links = {
                            self: `${this.channel.links.messages}/${args.item.data.sid}`,
                            conversation: this.channel.links.self,
                            messages_receipts: `${this.channel.links.messages}/${args.item.data.sid}/Receipts`,
                        };
                        const message$1 = new message.Message(args.item.index, args.item.data, this.channel, links, this.configuration, this.services);
                        if (this.messagesByIndex.has(message$1.index)) {
                            log.debug('Message arrived, but already known and ignored', this.channel.sid, message$1.index);
                            return;
                        }
                        this.messagesByIndex.set(message$1.index, message$1);
                        message$1.on('updated', (args) => this.emit('messageUpdated', args));
                        this.emit('messageAdded', message$1);
                    });
                    list.on('itemRemoved', args => {
                        log.debug(this.channel.sid + ' itemRemoved: ' + args.index);
                        let index = args.index;
                        if (this.messagesByIndex.has(index)) {
                            let message = this.messagesByIndex.get(index);
                            this.messagesByIndex.delete(message.index);
                            message.removeAllListeners('updated');
                            this.emit('messageRemoved', message);
                        }
                    });
                    list.on('itemUpdated', args => {
                        log.debug(this.channel.sid + ' itemUpdated: ' + args.item.index);
                        let message = this.messagesByIndex.get(args.item.index);
                        if (message) {
                            message._update(args.item.data);
                        }
                    });
                    return list;
                })
                    .catch(err => {
                    this.messagesListPromise = null;
                    if (this.services.syncClient.connectionState != 'disconnected') {
                        log.error('Failed to get messages object for channel', this.channel.sid, err);
                    }
                    log.debug('ERROR: Failed to get messages object for channel', this.channel.sid, err);
                    throw err;
                });
    }
    async unsubscribe() {
        if (this.messagesListPromise) {
            let entity = await this.messagesListPromise;
            entity.close();
            this.messagesListPromise = null;
        }
    }
    /**
     * Send Message to the channel
     * @param {String} message - Message to post
     * @param {any} attributes Message attributes
     * @returns Returns promise which can fail
     */
    async send(message, attributes = {}) {
        log.debug('Sending text message', message, attributes);
        return await this.services.commandExecutor.mutateResource('post', this.channel.links.messages, {
            body: message || '',
            attributes: attributes !== undefined ? JSON.stringify(attributes) : undefined,
        });
    }
    /**
     * Send Media Message to the channel
     * @param {FormData | Channel#SendMediaOptions} mediaContent - Media content to post
     * @param {any} attributes Message attributes
     * @returns Returns promise which can fail
     */
    async sendMedia(mediaContent, attributes = {}) {
        log.debug('Sending media message', mediaContent, attributes);
        let media;
        if (typeof FormData !== 'undefined' && (mediaContent instanceof FormData)) {
            log.debug('Sending media message as FormData', mediaContent, attributes);
            media = await this.services.mcsClient.postFormData(mediaContent);
        }
        else {
            log.debug('Sending media message as SendMediaOptions', mediaContent, attributes);
            let mediaOptions = mediaContent;
            if (!mediaOptions.contentType || !mediaOptions.media) {
                throw new Error('Media content <Channel#SendMediaOptions> must contain non-empty contentType and media');
            }
            media = await this.services.mcsClient.post(mediaOptions.contentType, mediaOptions.media, null);
        }
        return await this.services.commandExecutor.mutateResource('post', this.channel.links.messages, {
            media_sid: media.sid,
            attributes: attributes !== undefined ? JSON.stringify(attributes) : undefined
        });
    }
    /**
     * Returns messages from channel using paginator interface
     * @param {Number} [pageSize] Number of messages to return in single chunk. By default it's 30.
     * @param {String} [anchor] Most early message id which is already known, or 'end' by default
     * @param {String} [direction] Pagination order 'backwards' or 'forward', or 'forward' by default
     * @returns {Promise<Paginator<Message>>} last page of messages by default
     */
    async getMessages(pageSize, anchor, direction) {
        anchor = (typeof anchor !== 'undefined') ? anchor : 'end';
        direction = direction || 'backwards';
        return this._getMessages(pageSize, anchor, direction);
    }
    async wrapPaginator(order, page, op) {
        // We should swap next and prev page here, because of misfit of Sync and Chat paging conceptions
        let shouldReverse = order === 'desc';
        let np = () => page.nextPage().then(x => this.wrapPaginator(order, x, op));
        let pp = () => page.prevPage().then(x => this.wrapPaginator(order, x, op));
        return op(page.items).then(items => ({
            items: items.sort((x, y) => { return x.index - y.index; }),
            hasPrevPage: shouldReverse ? page.hasNextPage : page.hasPrevPage,
            hasNextPage: shouldReverse ? page.hasPrevPage : page.hasNextPage,
            prevPage: shouldReverse ? np : pp,
            nextPage: shouldReverse ? pp : np
        }));
    }
    _upsertMessage(index, value) {
        const cachedMessage = this.messagesByIndex.get(index);
        if (cachedMessage) {
            return cachedMessage;
        }
        const links = {
            self: `${this.channel.links.messages}/${value.sid}`,
            conversation: this.channel.links.self,
            messages_receipts: `${this.channel.links.messages}/${value.sid}/Receipts`,
        };
        const message$1 = new message.Message(index, value, this.channel, links, this.configuration, this.services);
        this.messagesByIndex.set(message$1.index, message$1);
        message$1.on('updated', (args) => this.emit('messageUpdated', args));
        return message$1;
    }
    /**
     * Returns last messages from channel
     * @param {Number} [pageSize] Number of messages to return in single chunk. By default it's 30.
     * @param {String} [anchor] Most early message id which is already known, or 'end' by default
     * @param {String} [direction] Pagination order 'backwards' or 'forward', or 'forward' by default
     * @returns {Promise<SyncPaginator<Message>>} last page of messages by default
     * @private
     */
    async _getMessages(pageSize, anchor, direction) {
        anchor = (typeof anchor !== 'undefined') ? anchor : 'end';
        pageSize = pageSize || 30;
        const order = direction === 'backwards' ? 'desc' : 'asc';
        return this.messagesListPromise
            .then(messagesList => messagesList.getItems({
            from: anchor !== 'end' ? anchor : void (0),
            pageSize,
            order,
            limit: pageSize // @todo Limit equals pageSize by default in Sync. This is probably not ideal.
        }))
            .then(page => this.wrapPaginator(order, page, items => Promise.all(items.map(item => this._upsertMessage(item.index, item.data)))));
    }
}

exports.Messages = Messages;
//# sourceMappingURL=messages.js.map
